(* $Id: c_check.ml,v 1.2 2006/08/01 10:31:12 deraugla Exp $ *)

open Rtdecl;
open Util;
open Xlib;

type check_event = [ CheckEvPress | CheckEvRelease ];

type check_args = unit;
type check_event_handler = widget -> check_event -> unit;

value check_band = 1;
value check_width = 9;

type check_local_info = { set : mutable bool };

value (check_local_info, get_check_local_info) =
  dynamo_local_info "check_local_info"
    (ref (None : option check_local_info))
;

value check_wsize att_val args xd =
  let w = check_width + 2 * (check_band + xd.motif_border) in
  let h = check_width + 2 * (check_band + xd.motif_border) in
  let b = 0 in
  {sh_width = w; sh_height = h; sh_border = b; base_width = w;
   base_height = h; width_inc = -1; height_inc = -1}
;

value check_wcreate att_val args xd pwin is_top in_popup wdesc x y wsh = (
  let mask = exposureMask lor buttonPressMask lor buttonReleaseMask in
  let info = check_local_info {set = False} in
  let win = create_window xd pwin is_top x y wsh att_val mask in
  Motif.motif_backg xd win;
  let wid = create_widget xd win is_top x y wsh wdesc info [] in
  add_widget att_val.name_att win wid
);

value check_expose xd wid inv =
  let size = max 1 (min wid.width wid.height) in
  let x = (wid.width - size) / 2 in
  let y = (wid.height - size) / 2 in
  Motif.motif_border xd wid.win x y size size
    (if inv then Motif.MBpush else Motif.MBpop)
;

value check_set wid val0 = (
  let li = get_check_local_info wid.info in
  let xd = wid.wid_xd in
  li.set := val0;
  check_expose xd wid val0
);

value check_val wid =
  let li = get_check_local_info wid.info in
  li.set
;

value check_wdispatch dir callb wid xev =
  let xd = wid.wid_xd in
  let li = get_check_local_info wid.info in
  let t = xEvent_type xev in
  if t = expose then check_expose xd wid li.set
  else if t = buttonPress then callb wid CheckEvPress
  else if t = buttonRelease then callb wid CheckEvRelease
  else ()
;

value check_desc attr args callb =
  let att_val = attribute_values attr in
  {wsize = check_wsize att_val args; wcreate = check_wcreate att_val args;
   wdestroy wid = remove_widget att_val.name_att wid.win wid;
   wdispatch = check_wdispatch args callb; wselect _ _ = ();
   wfreeze wid = (); wargs = no_args_info; filler = List.mem FillerAtt attr}
;
